create extension postgis
;

drop schema if exists ctrlcraft cascade
;

create schema ctrlcraft
;

-- the uuid package is not easilly available on windows, so we create our own
create sequence ctrlcraft.unique_name_seq
;

create or replace function ctrlcraft.unique_id()
returns varchar
language plpgsql volatile
as
$$
    begin
        return nextval('ctrlcraft.unique_name_seq')::varchar;
    end
$$
;

-- enum types
create type lithology AS ENUM ('Argile', 'Grès', 'Argile Oxydé', 'Grès Oxydé', 'ND')
;

-- table storing some settings
create table ctrlcraft.settings
(
  name varchar,
  value varchar
)
;

insert into ctrlcraft.settings(name, value) values('probe_interval', '20');
insert into ctrlcraft.settings(name, value) values('probe_min_date', ''); -- TODO: To be removed

-- table storing data from the ore smu file
create table ctrlcraft.smus
(
  id varchar primary key default ctrlcraft.unique_id(),
  smu_id varchar,
  deposit varchar(32),
  panel varchar(32),
  blast varchar(32),
  radiometry float,  -- u_krigee / quality value
  lithology lithology,
  percentage float,  -- recognition percentage for lithology
  unit varchar(32),
  geom geometry(PointZ, 32632)
)
;

create table ctrlcraft.truck_loads
(
  id varchar primary key default ctrlcraft.unique_id(),
  truck varchar,
  load_time timestamp,
  unload_time timestamp
)
;

-- table for all slabs
create table ctrlcraft.slabs
(
  id varchar primary key default ctrlcraft.unique_id(),
  smu varchar references ctrlcraft.smus(id),
  load varchar references ctrlcraft.truck_loads(id),
  deposit varchar(32),
  panel varchar(32),
  blast varchar(32),
  radiometry float,  -- u_krigee / quality value
  radiometry_probe float,  -- to update when a new probe data is added
  lithology lithology,
  unit varchar(32),
  geom geometry(MultiPolygonZ, 32632),
  centroid_z float, -- too long to retrieve from a MultiPolygonZ
  affectedradio varchar(32) default NULL -- when radiometry is updated by the CAN OP
)
;

create table ctrlcraft.trucks
(
  id varchar primary key,
  loads varchar[4]
)
;

create view ctrlcraft.slabs_available as
select * from ctrlcraft.slabs where radiometry is not null and load is null
;

create view ctrlcraft.truck_loads_status as
select count(*)::text,
  array_agg(s.id) as slabs,
  truck,
  deposit,
  blast,
  panel,
  avg(radiometry)::text as radiometry,
  avg(radiometry_probe)::text as radiometry_probe,
  COALESCE(to_char(unload_time, 'DD-MM-YYYY HH24:MI:SS'), '') as unload_time
from ctrlcraft.slabs s
join ctrlcraft.truck_loads l on s.load = l.id
where load is not null
group by truck,deposit,panel,blast,unload_time
order by unload_time
;

create table ctrlcraft.probe
(
  id varchar primary key default ctrlcraft.unique_id(),
  radiometry float,
  gammaup float,
  gammadown float,
  teneurup float,
  teneurdown float,
  time timestamp,
  geom geometry(Point, 32632)
)
;
CREATE INDEX IF NOT EXISTS probe_geom_idx ON ctrlcraft.probe USING gist (geom);

CREATE OR REPLACE VIEW ctrlcraft.probe_last_positions AS
(
	SELECT *
	FROM ctrlcraft.probe
	WHERE "time" >=
		(
			SELECT max("time") -
				(
					SELECT "value"
					FROM ctrlcraft.settings
					WHERE "name" = 'probe_interval'
				)::interval
			FROM ctrlcraft.probe
		)
	ORDER BY "time" DESC
);

CREATE TABLE ctrlcraft.fixed_measure(
id varchar(32) primary KEY default ctrlcraft.unique_id(),
slabid varchar(32),
radioup double precision,
radiolow double precision,
modedeg bool default false,
geom geometry(POINT, 32632)
);
CREATE INDEX IF NOT EXISTS fixed_measure_geom_idx ON ctrlcraft.fixed_measure USING gist (geom);

-- table storing co3 concentration from the Co3 file
create table ctrlcraft.co3
(
  deposit varchar(32),
  panel varchar(32),
  blast varchar(32),
  depth float,
  co3 float,
  geom geometry(PointZ, 32632)
)
;

-- table storing gate data
create table ctrlcraft.gate
(
  id varchar primary key default ctrlcraft.unique_id(),
  deposit varchar(32),
  panel varchar(32),
  blast varchar(32),
  truck varchar(32),
  time timestamp,
  radiometry float
)
;

-- function to load a truck with a slab
create function ctrlcraft.truck_load_slab(truckid varchar, wkt varchar)
RETURNS void
as
$$
DECLARE
  slabid varchar;
  loadid varchar;
BEGIN
  -- search the more above available slab at the given point (in wkt)
  slabid := (select id
             from ctrlcraft.slabs_available
             where st_intersects(geom, st_geomfromtext(wkt, 32632))
             order by centroid_z DESC limit 1);

  if slabid is not null
  then
    -- new load
    insert into ctrlcraft.truck_loads(truck, load_time) values(truckid, now())
    returning id into loadid;

    -- update truck with loads
    update ctrlcraft.trucks set loads = loads || loadid where id = truckid;

    -- update slab with load
    update ctrlcraft.slabs set load = loadid where id = slabid;
  end if;
END;
$$ language plpgsql
;

-- function to unload the last slab (whatever the truck)
create function ctrlcraft.truck_unload_slabs(truckid varchar)
RETURNS void
as
$$
DECLARE
BEGIN
  -- update departure time for loads
  update ctrlcraft.truck_loads set unload_time = now()
  where id in (select unnest(loads) from ctrlcraft.trucks where id = truckid);

  -- empty the trucks
  update ctrlcraft.trucks set loads = null where id = truckid;
END;
$$ language plpgsql
;

-- function to unload the last slab (whatever the truck)
create function ctrlcraft.truck_unload_last_slab()
RETURNS varchar
as
$$
DECLARE
  loadid varchar;
  truckid varchar;
BEGIN
  -- search the more recent inserted slab in trucks
  loadid := (select id from ctrlcraft.truck_loads
             where id in (select unnest(loads) from ctrlcraft.trucks)
             order by load_time DESC limit 1);

  if loadid is not null
  then
    -- update slab with load
    update ctrlcraft.slabs set load = null where load = loadid;

    -- remove load from trucks
    truckid := (select id from ctrlcraft.trucks where loadid = ANY(loads));
    update ctrlcraft.trucks set loads = array_remove(loads, loadid)
    where id = truckid;

    -- remove load
    delete from ctrlcraft.truck_loads where id = loadid;
  end if;

  return truckid;
END;
$$ language plpgsql
;

create function ctrlcraft.lithology_update(_lithology lithology, wkt varchar)
RETURNS void
as
$$
DECLARE
  slabid varchar;
BEGIN
  -- search the more above available slab at the given point (in wkt)
  slabid := (select id
             from ctrlcraft.slabs_available
             where st_intersects(geom, st_geomfromtext(wkt, 32632))
             order by centroid_z DESC limit 1);

  if slabid is not null
  then
    update ctrlcraft.slabs set lithology = _lithology where id = slabid;
  end if;
end;
$$ language plpgsql
;

CREATE FUNCTION ctrlcraft.add_fix_measure(_sondeUp int, _sondeLow int)
RETURNS void
  AS
$$
DECLARE
    avgradioup double precision;
    avgradiolow double precision;
    pt geometry(POINT, 32632);
    slabid varchar;
BEGIN
    -- get average radiometry and position from last 4 seconds
    SELECT
        AVG(teneurup) * _sondeUp, AVG(teneurdown) * _sondeLow,
        ST_Centroid (St_Collect (geom))
		INTO avgradioup, avgradiolow, pt
    FROM
        ctrlcraft.probe
    WHERE
        "time" >= (
            SELECT
                max("time") - interval '4' second
            FROM
                ctrlcraft.probe);
    -- search the more above available slab at the given point
    slabid := (
        SELECT
            id
        FROM
            ctrlcraft.slabs_available
        WHERE
            st_intersects (geom, pt)
        ORDER BY
            centroid_z DESC
        LIMIT 1);
    IF slabid IS NOT NULL THEN
      -- insert radiometry for the sonde
      INSERT INTO ctrlcraft.fixed_measure(slabid, radioup, radiolow, geom) VALUES (slabid, avgradioup, avgradiolow, pt);
      -- update radiometry for the slab
      PERFORM ctrlcraft.affect_radiometry_slab(slabid);
    END IF;
END;
$$
LANGUAGE plpgsql;

CREATE FUNCTION ctrlcraft.add_fix_measure_degraded(_teneurUp double precision, _teneurLow double precision, _wkt varchar )
RETURNS void
AS
$$
DECLARE
  slabid varchar;
BEGIN
  slabid := (
    SELECT
      id
    FROM
      ctrlcraft.slabs_available
    WHERE
      st_intersects(geom, st_geomfromtext(_wkt, 32632))
    ORDER BY
      centroid_z DESC
    LIMIT 1);

    IF slabid is NOT NULL THEN
      -- insert radiometry for the sonde
      INSERT INTO ctrlcraft.fixed_measure(slabid, radioup, radiolow, modedeg, geom) VALUES (slabid, _teneurUp, _teneurLow, true, st_geomfromtext(_wkt, 32632));
      -- update radiometry for the slab
      PERFORM ctrlcraft.affect_radiometry_slab(slabid);
    END IF;
END;
$$
LANGUAGE plpgsql;

CREATE FUNCTION ctrlcraft.affect_radiometry_slab(_slabid varchar)
    RETURNS void
    AS
$$
DECLARE
  nbRadioUp int;
  nbRadioLow int;
  sumRadioUp double precision;
  sumRadioLow double precision;
BEGIN
        SELECT SUM(radioup), sum(radiolow), count(*) FILTER (WHERE radioup > 0), count(*) FILTER (WHERE radiolow > 0)
        INTO sumRadioUp, sumRadioLow, nbRadioUp, nbRadioLow
        FROM ctrlcraft.fixed_measure
		WHERE
            slabid = _slabid;
        -- update slabs with mean radiometry
        UPDATE
            ctrlcraft.slabs
        SET
            radiometry = (sumRadioUp * nbRadioUp / (nbRadioUp + nbRadioLow)) + (sumRadioLow * nbRadioLow / (nbRadioUp + nbRadioLow))
			, affectedradio = nbRadioUp::varchar || '-' || nbRadioLow::varchar
        WHERE
            id = _slabid;
END;
$$
LANGUAGE plpgsql;

-- view with distinct deposits
create view ctrlcraft.deposits as
(
  select distinct deposit as id from ctrlcraft.smus order by id
)
;

-- view with distinct panels
create view ctrlcraft.panels as
(
  select distinct panel as id, deposit from ctrlcraft.smus order by id
)
;

-- view with distinct blast
create view ctrlcraft.blasts as
(
  select distinct blast as id, deposit, panel from ctrlcraft.smus
)
;

-- estimated blocs
create table ctrlcraft.estimated_blocs (
    sample_number integer,
    radiometry float,
    geom geometry(MultiPolygonZ, 32632)
)
;

create table ctrlcraft.estimated_blocs_stats (
  sample_number integer,
  deposit varchar(32),
  panel varchar(32),
  blast varchar(32),
  radiometry float  -- mean of the radiometry for all slabs
)
;

-- ---------------------------------------------------------------------------
--
-- Utility functions
--
-- ---------------------------------------------------------------------------
-- function to update probe radiometry for a slab
create function ctrlcraft.slab_update_probe_radiometry(
  _x double precision,
  _y double precision,
  _radiometry float
)
RETURNS setof void
as
$$
DECLARE
  point geometry;  -- Point
  mean_radiometry float;
  slab_id varchar;
BEGIN
  point := st_setsrid(st_point(_x, _y), 32632);
  slab_id := (select id from ctrlcraft.slabs_available
              where st_intersects(geom, point)
              order by centroid_z DESC limit 1);

  IF slab_id IS NOT NULL
  THEN
    mean_radiometry := (select radiometry_probe from ctrlcraft.slabs where id = slab_id);
    IF mean_radiometry IS NOT NULL
    THEN
      mean_radiometry := (mean_radiometry + _radiometry) / 2;
    ELSE
      mean_radiometry = _radiometry;
    END IF;
    update ctrlcraft.slabs set radiometry_probe = mean_radiometry where id = slab_id;
  END IF;
END;
$$ language plpgsql volatile
;

-- function to insert a smu
-- SMU dimension: 5 meters x 5 meters x 0.5 meters
-- slab dimension: 2.5 meters x 2.5 meters x 0.5 meters
create function ctrlcraft.insert_smu(
  _smuid text,
  _deposit text,
  _panel text,
  _blast text,
  _quality float,
  _lithology text,
  _percentage float,
  _unit text,
  _x double precision,
  _y double precision,
  _z float)
RETURNS setof void
as
$$
DECLARE
  smu_thick constant float := 0.5;  -- meters
  smu_side constant float := 5;  -- meters
  slab_side constant float := smu_side/2;
  slab0_centroid geometry;  -- PointZ
  slab0 geometry;  -- MultiPolygonZ
  slab1_centroid geometry;  -- PointZ
  slab1 geometry;  -- MultiPolygonZ
  slab2_centroid geometry;  -- PointZ
  slab2 geometry;  -- MultiPolygonZ
  slab3_centroid geometry;  -- PointZ
  slab3 geometry;  -- MultiPolygonZ
  lithology lithology;
  smu_id varchar;
BEGIN
  -- clean lithology text
  IF _lithology = 'argiles'
  THEN
    lithology = 'Argile';
  ELSIF _lithology = 'gres'
  THEN
    lithology = 'Grès';
  ELSIF _lithology = 'gres oxyde'
  THEN
    lithology = 'Grès Oxydé';
  ELSIF _lithology = 'argiles oxyde'
  THEN
    lithology = 'Argile Oxydé';
  ELSE
    lithology = 'ND';
  END IF;

  -- no value -> NULL in postgres
  IF _quality = '-999'
  THEN
    _quality = NULL;
  END IF;

  -- insert new SMU
  insert into ctrlcraft.smus(
    smu_id,
    deposit,
    panel,
    blast,
    radiometry,
    lithology,
    percentage,
    unit,
    geom
  ) values(
    _smuid,
    _deposit,
    _panel,
    _blast,
    _quality,
    lithology,
    _percentage,
    _unit,
    ST_SetSrid( ST_MakePoint( _x, _y, _z ), 32632 )
  )
  returning id into smu_id;

  -- build 3D slabs from SMU centroid
  slab0_centroid := ST_SetSrid( ST_MakePoint( _x - slab_side/2, _y + slab_side/2, _z ), 32632 );
  slab0 := ctrlcraft.create_3d_cube( slab0_centroid, slab_side, smu_thick );
  insert into ctrlcraft.slabs(smu, deposit, panel, blast, radiometry, lithology, unit, geom, centroid_z) values(smu_id, _deposit, _panel, _blast, _quality, lithology, _unit, slab0, _z);

  slab1_centroid := ST_SetSrid( ST_MakePoint( _x + slab_side/2, _y + slab_side/2, _z ), 32632 );
  slab1 := ctrlcraft.create_3d_cube( slab1_centroid, slab_side, smu_thick );
  insert into ctrlcraft.slabs(smu, deposit, panel, blast, radiometry, lithology, unit, geom, centroid_z) values(smu_id, _deposit, _panel, _blast, _quality, lithology, _unit, slab1, _z);

  slab2_centroid := ST_SetSrid( ST_MakePoint( _x + slab_side/2, _y - slab_side/2, _z ), 32632 );
  slab2 := ctrlcraft.create_3d_cube( slab2_centroid, slab_side, smu_thick );
  insert into ctrlcraft.slabs(smu, deposit, panel, blast, radiometry, lithology, unit, geom, centroid_z) values(smu_id, _deposit, _panel, _blast, _quality, lithology, _unit, slab2, _z);

  slab3_centroid := ST_SetSrid( ST_MakePoint( _x - slab_side/2, _y - slab_side/2, _z ), 32632 );
  slab3 := ctrlcraft.create_3d_cube( slab3_centroid, slab_side, smu_thick );
  insert into ctrlcraft.slabs(smu, deposit, panel, blast, radiometry, lithology, unit, geom, centroid_z) values(smu_id, _deposit, _panel, _blast, _quality, lithology, _unit, slab3, _z);
END;
$$ language plpgsql volatile
;

-- returns a centroid from a cube (z is not exact, but it's enough for our use
-- case)
create function ctrlcraft.cube_centroid(cube geometry)
RETURNS geometry
as
$$
DECLARE
  z double precision;
  centroid geometry;
BEGIN
  select AVG(zz) into z as avgz from ( select st_z(st_astext((st_dumppoints(cube)).geom)) as zz ) as zs;
  centroid := ST_MakePoint(ST_X(ST_Centroid(cube)), ST_Y(ST_Centroid(cube)), z);
  RETURN centroid;
END;
$$ language plpgsql volatile
;

-- function to build a 3D cube from its centroid
/*
       b _____________  c
        /|           /|
       / |          / |
      /  |         /  |
   a /___|________/ d |
     | F |________|___| g
     |   /        |   /
     |  /         |  /
     | /          | /
   e |/___________|/ h
*/
create function ctrlcraft.create_3d_cube(_centroid geometry, _side float, _thick float)
RETURNS geometry
as
$$
DECLARE
  a_x double precision;
  a_y double precision;
  a_z double precision;
  b_x double precision;
  b_y double precision;
  b_z double precision;
  c_x double precision;
  c_y double precision;
  c_z double precision;
  d_x double precision;
  d_y double precision;
  d_z double precision;
  e_x double precision;
  e_y double precision;
  e_z double precision;
  f_x double precision;
  f_y double precision;
  f_z double precision;
  g_x double precision;
  g_y double precision;
  g_z double precision;
  h_x double precision;
  h_y double precision;
  h_z double precision;
  geom_string text;
BEGIN

  a_x := ST_X(_centroid) - _side / 2;
  a_y := ST_Y(_centroid) - _side / 2;
  a_z := ST_Z(_centroid) + _thick / 2;

  b_x := ST_X(_centroid) - _side / 2;
  b_y := ST_Y(_centroid) + _side / 2;
  b_z := ST_Z(_centroid) + _thick / 2;

  c_x := ST_X(_centroid) + _side / 2;
  c_y := ST_Y(_centroid) + _side / 2;
  c_z := ST_Z(_centroid) + _thick / 2;

  d_x := ST_X(_centroid) + _side / 2;
  d_y := ST_Y(_centroid) - _side / 2;
  d_z := ST_Z(_centroid) + _thick / 2;

  e_x := a_x;
  e_y := a_y;
  e_z := ST_Z(_centroid) - _thick / 2;

  f_x := b_x;
  f_y := b_y;
  f_z := ST_Z(_centroid) - _thick / 2;

  g_x := c_x;
  g_y := c_y;
  g_z := ST_Z(_centroid) - _thick / 2;

  h_x := d_x;
  h_y := d_y;
  h_z := ST_Z(_centroid) - _thick / 2;

  /*
      a, b, c, d, a
      f, e, h, g, f
      e, a, d, h, e
      g, c, b, f, g
      f, b, a, e, f
      h, d, c, g, h
  */
  geom_string := FORMAT('MULTIPOLYGONZ(
                         ((%s %s %s, %s %s %s, %s %s %s, %s %s %s, %s %s %s)),
                         ((%s %s %s, %s %s %s, %s %s %s, %s %s %s, %s %s %s)),
                         ((%s %s %s, %s %s %s, %s %s %s, %s %s %s, %s %s %s)),
                         ((%s %s %s, %s %s %s, %s %s %s, %s %s %s, %s %s %s)),
                         ((%s %s %s, %s %s %s, %s %s %s, %s %s %s, %s %s %s)),
                         ((%s %s %s, %s %s %s, %s %s %s, %s %s %s, %s %s %s)) )',
                        a_x, a_y, a_z, b_x, b_y, b_z, c_x, c_y, c_z, d_x, d_y, d_z, a_x, a_y, a_z,
                        f_x, f_y, f_z, e_x, e_y, e_z, h_x, h_y, h_z, g_x, g_y, g_z, f_x, f_y, f_z,
                        e_x, e_y, e_z, a_x, a_y, a_z, d_x, d_y, d_z, h_x, h_y, h_z, e_x, e_y, e_z,
                        g_x, g_y, g_z, c_x, c_y, c_z, b_x, b_y, b_z, f_x, f_y, f_z, g_x, g_y, g_z,
                        f_x, f_y, f_z, b_x, b_y, b_z, a_x, a_y, a_z, e_x, e_y, e_z, f_x, f_y, f_z,
                        h_x, h_y, h_z, d_x, d_y, d_z, c_x, c_y, c_z, g_x, g_y, g_z, h_x, h_y, h_z);

  RETURN ST_GeomFromText(geom_string, 32632);
END;
$$ language plpgsql volatile
;

-- function to insert a panel
-- SMU dimension: 25 meters x 25 meters x 0.5 meters
create function ctrlcraft.insert_estimated_bloc(
  _sample_number integer,
  _radiometry float,
  _x double precision,
  _y double precision,
  _z float)
RETURNS setof void
as
$$
DECLARE
  panel_side constant float := 25.0;  -- meters
  panel_thick constant float := 0.5;  -- meters
  geom geometry;  -- MultiPolygonZ
  geom_centroid geometry;  -- PointZ
BEGIN
  -- no value -> NULL in postgres
  IF _radiometry = '-999'
  THEN
    _radiometry = NULL;
  END IF;

  geom_centroid := ST_SetSrid( ST_MakePoint( _x, _y, _z ), 32632 );
  geom := ctrlcraft.create_3d_cube( geom_centroid, panel_side, panel_thick );

  insert into ctrlcraft.estimated_blocs values(_sample_number, _radiometry, geom);
END;
$$ language plpgsql volatile
;

create function ctrlcraft.estimated_blocs_stats_compute()
RETURNS void
as
$$
DECLARE
  blast record;
  res record;
BEGIN
  FOR blast in select id, deposit, panel from ctrlcraft.blasts
  LOOP
    for res in select avg(s.radiometry) as rad_mean, ms.sample_number
    from ctrlcraft.slabs_available s
    join ctrlcraft.estimated_blocs ms on ST_3DIntersects(s.geom, ms.geom)
    where s.panel = blast.panel and s.blast = blast.id
    group by ms.sample_number, ms.radiometry
    order by ms.sample_number
    LOOP
      insert into ctrlcraft.estimated_blocs_stats
      values( res.sample_number, blast.deposit, blast.panel, blast.id, res.rad_mean);
    END LOOP;
  END LOOP;
END;
$$ language plpgsql volatile
;

create function ctrlcraft.probe_new_data(
  _date varchar,
  _x double precision,
  _y double precision,
  _radiometry float,
  _gammaup float,
  _gammadown float,
  _teneurup float,
  _teneurdown float)
RETURNS setof void
as
$$
DECLARE
  geom geometry;
  t timestamp;
BEGIN
  geom := ST_SetSrid( ST_MakePoint( _x, _y ), 32632 );
  t := to_timestamp(_date, 'DD/MM/YYYY HH24:MI:SS') at time zone 'Etc/UTC';
  insert into ctrlcraft.probe(radiometry, time, geom, gammaup, gammadown, teneurup, teneurdown)
  values( _radiometry, t, geom, _gammaup, _gammadown, _teneurup, _teneurdown );
  perform ctrlcraft.slab_update_probe_radiometry(_x, _y, _radiometry);
END;
$$ language plpgsql volatile
;

create function ctrlcraft.probe_to_csv(
  _filename varchar
)
RETURNS setof void
as
$$
DECLARE
BEGIN
  execute 'copy (select radiometry, time, st_astext(geom) from ctrlcraft.probe) to ''' || _filename || ''' with csv;';
END;
$$ language plpgsql volatile
;
